# The keywords for the different sections of the file
RELATION_SECTION_KEYWORD = '@relation'
ATTRIBUTE_SECTION_KEYWORD = '@attribute'
DATA_SECTION_KEYWORD = '@data'

# Keywords for the different attribute datatypes
NUMERIC_ATTRIBUTE_KEYWORD = 'numeric'
INTEGER_ATTRIBUTE_KEYWORD = 'integer'
REAL_ATTRIBUTE_KEYWORD = 'real'
NOMINAL_ATTRIBUTE_KEYWORD = 'nominal'
DATE_ATTRIBUTE_KEYWORD = 'date'
STRING_ATTRIBUTE_KEYWORD = 'string'
RELATIONAL_ATTRIBUTE_KEYWORD = 'relational'

# The symbol that represents a missing value
MISSING_VALUE_SYMBOL = '?'

# The symbol that represents the beginning of a comment
COMMENT_SYMBOL = '%'
